import re
from typing import Iterator

from rich.console import Console
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    TaskID,
    TaskProgressColumn,
    TextColumn,
)
from rich.style import Style
from rich.syntax import Syntax

from gl_search.models import RepoResult


class MatchFinder:
    match_position: list[tuple[int, int]]

    def __init__(self, text: str, search_code_input: str) -> None:
        self.match_position = []
        self._process(text, search_code_input)

    def _process(self, text: str, search_code_input: str) -> None:
        splitted_lines = text.split("\n")

        accumulator = 0

        for line, splitted_line in enumerate(splitted_lines, start=1):
            accumulator += len(splitted_line)

            regex_findall_result: Iterator[re.Match[str]] = re.finditer(
                rf"{search_code_input}", splitted_line, flags=re.IGNORECASE
            )

            for match in regex_findall_result:
                self.match_position.append(((line, match.start()), (line, match.end())))


class ProcessUserFeedback:
    SEARCHING_GROUPS = "Searching groups"
    SEARCHING_REPOS = "Searching repos"
    SEARCHING_CODE = "Searching code"
    progress: Progress
    _tasks: dict[str, TaskID]

    def __init__(self):
        self.progress = Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            MofNCompleteColumn(),
        )

        self._tasks = dict()
        self._add_tasks()

    def _get_task(self, task_name: str) -> TaskID:
        return self._tasks[task_name]

    def set_completed(self, task_name: str) -> None:
        self.progress.update(self._get_task(task_name), completed=True)

    def set_total(self, task_name: str, new_total: float) -> None:
        self.progress.update(self._get_task(task_name), total=new_total)

    def set_advance(self, task_name: str) -> None:
        self.progress.advance(self._get_task(task_name))

    def set_visible(self, task_name: str) -> None:
        self.progress.update(self._get_task(task_name), visible=True)

    def _add_tasks(self) -> None:
        self._tasks[self.SEARCHING_GROUPS] = self.progress.add_task(self.SEARCHING_GROUPS, total=1)
        self._tasks[self.SEARCHING_REPOS] = self.progress.add_task(
            self.SEARCHING_REPOS, total=100, visible=False
        )
        self._tasks[self.SEARCHING_CODE] = self.progress.add_task(
            self.SEARCHING_CODE, total=100, visible=False
        )


process_user_feedback = ProcessUserFeedback()


def print_results(results: list[RepoResult], search_code_input: str) -> None:
    console = Console()

    for entry in results:
        if not entry.results:
            continue

        console.print(f"Proj : {entry.name}\n")

        for content in entry.results:
            match_finder = MatchFinder(content.data, search_code_input)

            text = Syntax(
                content.data, content.path.split(".")[-1], start_line=content.start_line, line_numbers=True
            )

            for match_position in match_finder.match_position:
                text.stylize_range(Style(bgcolor="deep_pink4"), *match_position)

            console.print(f"{content.filename} - {entry.web_url}/-/blob/{content.ref}/{content.path}\n")
            console.print(text)
            console.print("-----------\n")



from typing import Any

import click
import toml

from gl_search.config import BLOCK_SETTINGS_NAME, SETTINGS_FILE_PATH


@click.group("settings")
def config_cli() -> None: ...


def _load_settings(file_path: str) -> dict[str, Any]:
    try:
        current_settings = toml.load(file_path)
    except FileNotFoundError:
        current_settings = {BLOCK_SETTINGS_NAME: {}}

    return current_settings


def _write_settings(file_path: str, current_settings: dict[str, Any]) -> None:
    with open(file_path, "w") as file:
        toml.dump(current_settings, file)


@config_cli.command(name="setup-token", short_help="Command to register the token to be used with requests.")
@click.argument("token")
def setup_token(token: str) -> None:
    """This command register the TOKEN that going to be used with requests.

    The TOKEN is generate on your gitlab profile [1] with scope read_api.

    [1] https://gitlab.com/-/profile/personal_access_tokens

    """

    current_settings = _load_settings(SETTINGS_FILE_PATH)
    current_settings[BLOCK_SETTINGS_NAME].update({"GITLAB_PRIVATE_TOKEN": token})

    _write_settings(SETTINGS_FILE_PATH, current_settings)

    click.secho("Successfully registered.", fg="green")


@config_cli.command(
    name="setup-gitlab-address", short_help="Command to register the gitlab address to be used with requests."
)
@click.argument("address")
def setup_gitlab_address(address: str) -> None:
    """This command register the address that going to be used with requests."""

    current_settings = _load_settings(SETTINGS_FILE_PATH)
    current_settings[BLOCK_SETTINGS_NAME].update({"GITLAB_URL": address})

    _write_settings(SETTINGS_FILE_PATH, current_settings)

    click.secho("Successfully registered.", fg="green")



import logging
import time
from http import HTTPStatus
from random import uniform
from typing import Callable

import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from .config import settings
from .display import process_user_feedback
from .models import RequestDescribe

logger = logging.getLogger(__name__)

retries = Retry(
    total=3,
    backoff_factor=1,
    status_forcelist=[
        HTTPStatus.TOO_MANY_REQUESTS,
        HTTPStatus.BAD_GATEWAY,
        HTTPStatus.SERVICE_UNAVAILABLE,
        HTTPStatus.GATEWAY_TIMEOUT,
    ],
    raise_on_status=False,
)

request_session = requests.Session()

request_session.mount("http://", HTTPAdapter(max_retries=retries))
request_session.mount("https://", HTTPAdapter(max_retries=retries))

request_session.headers.update({"PRIVATE-TOKEN": settings.GITLAB_PRIVATE_TOKEN})


def retrieve_data(
    request: RequestDescribe,
    transform_data: Callable = lambda value: value,
    max_random_time_for_sleep: float = 0,
) -> list:
    count: int = 0
    data_list: list[dict] = []
    while True or count < settings.MAX_DEEP_SEARCH:
        count += 1

        if logger.isEnabledFor(logging.DEBUG):
            process_user_feedback.progress.log("URL: {} PARAMS: {}".format(request.url, request.params))

        response: requests.Response = request_session.get(request.url, params=request.params, timeout=10)

        if response.status_code == HTTPStatus.TOO_MANY_REQUESTS:
            process_user_feedback.progress.print(
                "URL: {} PARAMS: {} HttpStatus Code 429 SKIP this search".format(request.url, request.params)
            )
            break
        elif not response.status_code == HTTPStatus.OK:
            raise Exception(f"invalid status code {response.status_code}")

        data_list.extend([transform_data(data) for data in response.json()])

        try:
            request.url: str = response.links["next"]["url"]
            request.params = {}
        except KeyError:
            break

        time.sleep(uniform(0, max_random_time_for_sleep))

    return data_list



import concurrent.futures
from asyncio import Future

from .config import settings
from .display import process_user_feedback
from .models import Repo, RepoResult, RequestDescribe, SearchEntryResult, SearchParams, SearchRepoParams
from .utils import retrieve_data


def _retrieve_groups_ids(params: SearchParams) -> list[int]:
    request = RequestDescribe(
        url=f"{settings.GITLAB_URL}/api/v4/groups",
        params={
            "per_page": "100",
            "pagination": "keyset",
            "order_by": "id",
            "sort": "asc",
        },
    )

    data_list: list[int] = retrieve_data(
        request, lambda value: value["id"], max_random_time_for_sleep=params.max_random_time_for_sleep
    )

    return data_list


def _retrieve_repositories_by(group_id: int, params: SearchParams) -> set[Repo]:
    request = RequestDescribe(
        url=f"{settings.GITLAB_URL}/api/v4/groups/{group_id}/projects",
        params={
            "per_page": "100",
            "include_subgroups": "true",
        },
    )
    repos = set(
        retrieve_data(
            request, lambda value: Repo(**value), max_random_time_for_sleep=params.max_random_time_for_sleep
        )
    )

    return {repo for repo in repos if repo.visibility in params.visibility}


def _retrieve_information_from_repositories_of_each_group(
    groups_id: list[int], params: SearchParams
) -> set[Repo]:
    process_user_feedback.set_total(process_user_feedback.SEARCHING_REPOS, len(groups_id))
    process_user_feedback.set_visible(process_user_feedback.SEARCHING_REPOS)
    repos: set[Repo] = set()

    with concurrent.futures.ThreadPoolExecutor(max_workers=params.max_workers) as executor:
        futures: dict[Future[Repo], int] = {
            executor.submit(_retrieve_repositories_by, group_id, params): group_id for group_id in groups_id
        }
        for future in concurrent.futures.as_completed(futures):
            data: set[Repo] = future.result()
            repos.update(data)
            process_user_feedback.set_advance(process_user_feedback.SEARCHING_REPOS)

    return repos


def _search_in_repo(search_params: SearchRepoParams) -> list[SearchEntryResult]:
    params = {"scope": "blobs", "search": search_params.search_with_params, "per_page": 100}

    request = RequestDescribe(
        url=f"{settings.GITLAB_URL}/api/v4/projects/{search_params.repo_id}/search",
        params=params,
    )
    data_list: list[SearchEntryResult] = retrieve_data(
        request,
        lambda value: SearchEntryResult(**value),
        max_random_time_for_sleep=search_params.max_random_time_for_sleep,
    )

    return data_list


def _search_code(
    repos: list[Repo],
    params: SearchParams,
) -> list[RepoResult]:
    process_user_feedback.set_total(process_user_feedback.SEARCHING_CODE, len(repos))
    process_user_feedback.set_visible(process_user_feedback.SEARCHING_CODE)
    results: list[RepoResult] = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=params.max_workers) as executor:
        futures: dict[Future[list[SearchEntryResult]], Repo] = {
            executor.submit(
                _search_in_repo,
                SearchRepoParams(
                    repo_id=repo.id,
                    **params.dict(),
                ),
            ): repo
            for repo in repos
        }
        for future in concurrent.futures.as_completed(futures):
            repo: Repo = futures[future]

            data: list[SearchEntryResult] = future.result()

            results.append(RepoResult(name=repo.name, web_url=repo.web_url, results=data))

            process_user_feedback.set_advance(process_user_feedback.SEARCHING_CODE)

    return results


def _retrieve_groups(params: SearchParams) -> list[int]:
    if params.groups:
        groups_ids = list(map(int, params.groups.split(",")))
    else:
        groups_ids = _retrieve_groups_ids(params)

    process_user_feedback.set_completed(process_user_feedback.SEARCHING_GROUPS)
    return groups_ids


def search(params: SearchParams) -> list[dict[str, str]]:
    groups_ids = _retrieve_groups(params)

    repos = _retrieve_information_from_repositories_of_each_group(groups_ids, params)

    result = _search_code(repos, params)

    return result



import click
from click import CommandCollection
from dynaconf.validator import ValidationError

from .clis import config_cli, search_cli
from .config import settings


class CustomCommandCollection(CommandCollection):
    def invoke(self, ctx):
        args = [*ctx.protected_args, *ctx.args]
        cmd_name, cmd, args = self.resolve_command(ctx, args)

        COMMANDS_THAT_NEED_THE_LIBRARY_CONFIGURED = ["search"]

        if cmd_name in COMMANDS_THAT_NEED_THE_LIBRARY_CONFIGURED:
            try:
                settings.validators.validate_all()
            except ValidationError as error:
                click.secho(f"Settings error: {error}", fg="red")
                ctx.exit()
        return super().invoke(ctx)


cli = CustomCommandCollection(name="gl-search", sources=[search_cli, config_cli])
